/*
 * dsp.c: /dev/dsp access code
 *
 * This file is part of GTick
 *
 *
 * Copyright (c) 1999, Alex Roberts
 * Copyright (c) 2003, Roland Stigge <stigge@antcom.de>
 *
 * GTick is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * GTick is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GTick; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
 
/* GNU headers */
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* OSS headers */
#ifdef FREEBSD
#include <machine/soundcard.h>
#else
#include <linux/soundcard.h>
#endif

/* GTK+ headers */
#include <glib.h>

/* own headers */
#include "g711.h"
#include "globals.h"
#include "metro.h"
#include "dsp.h"
#include "tickdata.c"

/* time to write ahead at least in milliseconds */
#define WRITE_AHEAD_INTERVAL 100

/* default DSP settings */
#define DEFAULT_RATE 44100
#define DEFAULT_FORMAT AFMT_S16_LE
#define DEFAULT_CHANNELS 1

typedef struct format_t {
  int format;
  char* name;
  int samplesize;
  char* description;
} format_t;

static format_t formats[] = {
  { AFMT_MU_LAW,    "AFMT_MU_LAW",     8, "8 bit mu-law"                     },
  { AFMT_A_LAW,     "AFMT_A_LAW",      8, "8 bit A-law"                      },
  { AFMT_IMA_ADPCM, "AFMT_IMA_ADPCM",  4, "4 bit IMA ADPCM"                  },
  { AFMT_U8,        "AFMT_U8",         8, "unsigned, 8 bit"                  },
  { AFMT_S16_LE,    "AFMT_S16_LE",    16, "signed, 16 bit (little endian)"   },
  { AFMT_S16_BE,    "AFMT_S16_BE",    16, "signed, 16 bit (big endian)"      },
  /* equals AFMT_S16_LE or AFMT_S16_BE, shouldn't be reported: */
  { AFMT_S16_NE,    "AFMT_S16_NE",    16, "signed, 16 bit (local CPU endian)"},
  /* not supported by default soundcard.h
  { AFMT_S32_LE,    "AFMT_S32_LE",    32, "signed, 32 bit (little endian)"   },
  { AFMT_S32_BE,    "AFMT_S32_BE",    32, "signed, 32 bit (big endian)"      },
  */
  { AFMT_U16_LE,    "AFMT_U16_LE",    16, "unsigned, 16 bit (little endian)" },
  { AFMT_U16_BE,    "AFMT_U16_BE",    16, "unsigned, 16 bit (big endian)"    },

  { 0,              "unknown",        0,  "???"                              }
};

/*
 * Writes the specified <sample> (16 bit signed) to raw buffer <dest> with
 * specified <format> (see formats, above)
 */
static void gtEncodeSample(short sample, int format, unsigned char* dest)
{
  static int error = 0;
    
  switch (format) {
  case AFMT_MU_LAW:
    *dest = linear2ulaw(sample);
    break;
  case AFMT_A_LAW:
    *dest = linear2alaw(sample);
    break;
  case AFMT_IMA_ADPCM:
    if (!error) {
      fprintf(stderr, "NOTE: Can't generate samples due to still unsupported "
	              "format (AFMT_IMA_ADPCM).\n");
      error = 1;
    }
    *dest = 0;
    break;
  case AFMT_U8:
    *dest = (unsigned char) (sample / 256 + 128);
    break;
  case AFMT_S16_LE:
    *dest = (unsigned char) (sample & 0xff);
    *(dest + 1) = (unsigned char) (sample >> 8 & 0xff);
    break;
  case AFMT_S16_BE:
    *dest = (unsigned char) (sample >> 8 & 0xff);
    *(dest + 1) = (unsigned char) (sample & 0xff);
    break;
  case AFMT_U16_LE:
    *dest = (unsigned char) ((sample ^ 0x8000) & 0xff);
    *(dest + 1) = (unsigned char) ((sample ^ 0x8000) >> 8 & 0xff);
    break;
  case AFMT_U16_BE:
    *dest = (unsigned char) ((sample ^ 0x8000) >> 8 & 0xff);
    *(dest + 1) = (unsigned char) ((sample ^ 0x8000) & 0xff);
    break;
  default:
    if (!error) {
      fprintf(stderr,
	      "NOTE: Can't generate samples due to unsupported format.\n");
      error = 1;
    }
  }
}

/*
 * returns x if x < limit, else (limit - 1); thus it returns an exclusively
 * limited value
 */
static int gtLimit(int x, int limit) {
  if (x < limit)
    return x;
  else
    return limit - 1;
}

/*
 * Generates Audio data fit for playback on initialized DSP parameters of
 * which are stored in metro
 *
 * input:
 *     from:          the sample data source in signed 16 bit format
 *     from_size:     number of frames in <from>
 *     from_rate:     the rate of <from>
 *     from_channels: number of channels in <from>
 *     metro:         the metro_t structure holding rate, channels, samplesize
 *                    and format of the initialized dsp
 *     
 * output:
 *     to:            the pointer to the allocated data, ready for playback
 *     return value:  number of bytes generated, -1 on error
 *
 * NOTES:
 *  - If fromchannels equals the number of channels of playback device,
 *    no mixing is performed. Else, the input is mixed down to 1 channel
 *    which is equally directed to all output channels
 *  - <to> will be allocated by gtGenerate, but has to be free()d by caller
 */
static int gtGenerate(short* from,
                      int from_size, int from_rate, int from_channels,
                      metro_t* metro, unsigned char** to)
{
  double speed_factor = (double) from_rate / metro->rate; /* output to input */
  int result_size = from_size * metro->rate / from_rate; /* number of frames */
  unsigned char* result;
  int i, j;

  if (!(result = (unsigned char*)
	malloc(result_size * metro->samplesize / 8 * metro->channels)))
  {
    fprintf(stderr, "Error: Couldn't allocate playback output buffer.\n");
    return -1;
  }

  for (i = 0; i < result_size; i++) { /* for each output frame */
    double mixdown = 0.0; /* only needed if metro->channels != from_channels */

    if (metro->channels != from_channels) { /* mixdown */
      double leftbound = i * speed_factor;
      double rightbound = (i + 1) * speed_factor;

      while (leftbound < rightbound) {
	int index = (int)leftbound;
	double dummy;
	double frac = modf(leftbound, &dummy);
        double weight;
	
	if (rightbound - leftbound < 1)
	  weight = rightbound - leftbound;
	else
	  weight = 1.0;
	
        for (j = 0; j < from_channels; j++) {
	  mixdown += weight * (
	      (1.0 - frac) * from[from_channels * index + j] +
	      frac * from[from_channels * gtLimit(index + 1, from_size) + j]);
        }

	leftbound += 1.0;
      }
      mixdown /= (speed_factor * from_channels);
    }
    for (j = 0; j < metro->channels; j++) { /* for each channel */
      double sample;
      
      if (metro->channels != from_channels) {
	sample = mixdown;
      } else {
        double leftbound = i * speed_factor;
        double rightbound = (i + 1) * speed_factor;

	sample = 0.0;
        while (leftbound < rightbound) {
	  int index = (int)leftbound;
	  double dummy;
	  double frac = modf(leftbound, &dummy);
          double weight;
	
	  if (rightbound - leftbound < 1)
	    weight = rightbound - leftbound;
	  else
	    weight = 1.0;
	
	  sample += weight * (
	      (1.0 - frac) * from[from_channels * index + j] +
	      frac * from[from_channels * gtLimit(index + 1, from_size) + j]);

	  leftbound += 1.0;
        }
        sample /= speed_factor;
      }

      gtEncodeSample((short) sample, metro->format,
          &result[(i * metro->channels + j) * metro->samplesize / 8]);

    }
  }

  *to = result;
  return result_size * metro->channels * metro->samplesize / 8;
}

/*
 * Open DSP and prepare metronome <metro> to play
 */
int gtInitDSP (metro_t *metro)
{
  unsigned int i;
  unsigned int format_index;
  int requested_format = DEFAULT_FORMAT;
  audio_buf_info info;
  short silencelevel = 0;
  short *tmp_buf; /* temporary buffer for generation of different sounds */
  
  metro->fragmentsize = 0x7fff0008; /* at least request fragment size 2^8=256 */
                                    /* = minimum recommended size */
  if (debug)
    g_print ("gtInitDSP: Initialising /dev/dsp..\n");

  /* Initialise /dev/dsp */
  if ((metro->dspfd = open ("/dev/dsp", O_WRONLY)) == -1)
    {
      perror ("/dev/dsp");
      return -1;
    }

  if (ioctl(metro->dspfd, SNDCTL_DSP_SETFRAGMENT, &metro->fragmentsize) == -1) {
    perror("SNDCTL_DSP_SETFRAGMENT");
  }

  /* Query driver for supported formats for debugging convenience */
  if (debug) {
    int mask;
    if (ioctl(metro->dspfd, SNDCTL_DSP_GETFMTS, &mask) == -1) {
      perror("SNDCTL_DSP_GETFMTS");
      exit(1);
    }
    g_print("Supported formats:\n");
    for (i = 0; i < sizeof(formats) / sizeof(format_t); i++) {
      if (mask & formats[i].format)
	g_print("  %s (%s)\n", formats[i].name, formats[i].description);
    }
  }

  /* set up output format */
  metro->format = requested_format;
  if (ioctl (metro->dspfd, SNDCTL_DSP_SETFMT, &metro->format) == -1)
    { /* Fatal error */
      perror ("SNDCTL_DSP_SETFMT");
      exit(1);
    }
  
  for (format_index = 0;
    formats[format_index].format != 0 &&
      formats[format_index].format != metro->format;
    format_index++);
  
  if (debug) {
    g_print("gtInitDSP: Used sample format: %s (%s)\n",
	    formats[format_index].name, formats[format_index].description);
  }

  metro->samplesize = formats[format_index].samplesize;
		  
  /* Set dsp to default: mono */
  metro->channels = DEFAULT_CHANNELS;
  if (ioctl (metro->dspfd, SNDCTL_DSP_CHANNELS, &metro->channels) == -1) {
    perror("SNDCTL_DSP_CHANNELS");
    exit(1);
  }
  if (debug) {
    g_print("gtInitDSP: Number of channels = %d\n", metro->channels);
  }
  
  /* Set the DSP rate (in Hz) */
  metro->rate = DEFAULT_RATE; /* requested default */
  if (ioctl (metro->dspfd, SNDCTL_DSP_SPEED, &metro->rate) == -1) {
    perror("SNDCTL_DSP_SPEED");
    exit(1);
  }
  if (debug) {
    g_print("gtInitDSP: Sampling rate = %d\n", metro->rate);
  }

  /* "verify" fragment size:
  *  let the driver actually calculate the fragment size
  */
  if (ioctl(metro->dspfd, SNDCTL_DSP_GETBLKSIZE, &metro->fragmentsize)) {
    perror("SNDCTL_DSP_GETBLKSIZE");
  }

  if (debug)
    g_print ("gtInitDSP: fragment size = %d\n", metro->fragmentsize);

  if (!(metro->fragment = g_malloc(metro->fragmentsize))) {
    perror("g_malloc");
  }

  /* get total number of fragments in dsp buffer */
  if (ioctl(metro->dspfd, SNDCTL_DSP_GETOSPACE, &info) == -1) {
    perror("SNDCTL_DSP_GETOSPACE");
  }
  metro->fragstotal = info.fragstotal;
  if (debug)
    g_print("gtInitDSP: total number of fragments in DSP buffer = %d\n",
	    metro->fragstotal);
 
  /* silence */
  gtGenerate(&silencelevel, 1, metro->rate, 1, metro, &metro->silence);

  /* generate single ticks */
  metro->td0_size = gtGenerate(tickdata, sizeof(tickdata) / sizeof(short),
                               44100, 1, metro, &metro->tickdata0);

  /* generate first tick */
  metro->td1_size = gtGenerate(tickdata, sizeof(tickdata) / sizeof(short),
                               88200, 1, metro, &metro->tickdata1);

  /* generate secondary ticks */
  if (!(tmp_buf = (short*) malloc(sizeof(tickdata)))) {
    perror("malloc");
  }
  for (i = 0; i < sizeof(tickdata) / sizeof(short); i++) {
    tmp_buf[i] = tickdata[i] / 2;
  }
  metro->td2_size = gtGenerate(tmp_buf, sizeof(tickdata) / sizeof(short),
                               44100, 1, metro, &metro->tickdata2);
  free(tmp_buf);

  metro->cyclepos = 0; /* init */

  metro->sound_event_id = gtk_timeout_add(WRITE_AHEAD_INTERVAL / 2,
                                          gtDSPFeed,
					  metro);

  return 0;
}

/*
 * close device and clean up
 */
void gtCloseDSP(metro_t* metro)
{
  gtk_timeout_remove(metro->sound_event_id);
  
  if (ioctl(metro->dspfd, SNDCTL_DSP_RESET, 0) == -1) {
    perror("SNDCTL_DSP_RESET");
  }
  
  g_free(metro->fragment);
  g_free(metro->tickdata0);
  g_free(metro->tickdata1);
  g_free(metro->tickdata2);
  g_free(metro->silence);

  if (debug)
    g_print ("gtCloseDSP: Closing /dev/dsp..\n");
  close (metro->dspfd);
}

/*
 * Feed dsp device with next samples
 *
 * used as output start and callback
 */
gboolean gtDSPFeed(gpointer data)
{
  metro_t *metro = (metro_t *)data;

  audio_buf_info info; /* OSS structure to obtain buffering parameters */
  int ticklen = (metro->rate * metro->period / 1000) *
                metro->channels * metro->samplesize / 8;
  int cyclelen = metro->meter * ticklen;
  unsigned char *the_data; /* pointer to actual buffer */
  int data_size;           /* size of actual buffer */

  int fragments; /* number of fragments yet to write */
  int limit; /* number of fragments we want to have filled */
 
  /* get number of fragments to write to dsp */
  if (ioctl(metro->dspfd, SNDCTL_DSP_GETOSPACE, &info) == -1) {
    perror("SNDCTL_DSP_GETOSPACE");
  }

  limit =
    metro->rate * metro->channels * metro->samplesize / 8 *
    WRITE_AHEAD_INTERVAL / (1000 * metro->fragmentsize);
  if (limit < 2) /* we want to have filled at least 2 fragments */
    limit = 2;
  fragments = limit - (info.fragstotal - info.fragments);

  if (metro->cyclepos >= cyclelen)
    metro->cyclepos = 0;

  /* write as many fragments as possible */
  while (fragments > 0) {
    int i;
    
    /* generate fragment */
    for (i = 0; i < metro->fragmentsize; i++) {
      if (metro->ticking) { /* metronome on */
        int index;
	
	if (metro->meter == 1) {                 /* single ticks */
	  the_data = metro->tickdata0;
	  data_size = metro->td0_size;
	} else if (metro->cyclepos / ticklen) {  /* sound of 2nd tick */
	  the_data = metro->tickdata2;
	  data_size = metro->td2_size;
	} else {                                 /* accentuate 1st tick */
	  the_data = metro->tickdata1;
	  data_size = metro->td1_size;
	}

	if ((index = metro->cyclepos % ticklen) < data_size) { /* tick! */
	  metro->fragment[i] = the_data[index];
	} else { /* silence (between ticks)  */
	  metro->fragment[i] =
	    metro->silence[i % (metro->samplesize / 8 * metro->channels)];
	}

        metro->cyclepos ++;
        if (metro->cyclepos >= cyclelen)
	  metro->cyclepos = 0;
	
      } else { /* silence (metronome off) */
        metro->fragment[i] =
	  metro->silence[i % (metro->samplesize / 8 * metro->channels)];
      }
    }

    write(metro->dspfd, metro->fragment, metro->fragmentsize);
    fragments--;
  }

  return TRUE;
}

